local Utopia = LibStub("AceAddon-3.0"):GetAddon("Utopia", true)
if (not Utopia or Utopia.modules.UpTime) then
	return
end
Utopia:UpdateVersion("$Revision$")
local function d(...)
	ChatFrame1:AddMessage("|cFFFF8080DEBUG:|r "..format(...), 0.8, 0.8, 0.8)
end

local module = Utopia:NewModule("UpTime", "AceHook-3.0", "AceEvent-3.0", "AceTimer-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Utopia")
local gui = LibStub("AceGUI-3.0")
local bossid = LibStub("LibBossIDs-1.0").BossIDs
local db, dlookup, blookup, mlookup
local active
local new, del, copy, deepDel = Utopia.new, Utopia.del, Utopia.copy, Utopia.deepDel
local Gradient, UnitFullName, ClassColour, SmoothColour = Utopia.Gradient, Utopia.UnitFullName, Utopia.ClassColour, Utopia.SmoothColour
local rotate, propercase = Utopia.rotate, Utopia.propercase
local band = bit.band

local BossIDWeakLookup = setmetatable({}, {
	__mode = "k",
	__index = function(self, guid)
		local id = tonumber(strsub(guid, -12, -7), 16)
		self[guid] = bossid[id] and true or false
		return self[guid]
	end,
})

module.options = {
	type = "group",
	name = L["Up-Time"],
	desc = L["Records debuff up-time on fights"],
	get = function(info) return module.db.profile[info[#info]] end,
	set = function(info, value) module.db.profile[info[#info]] = value end,
	guiInline = true,
	order = 100,
	args = {
		bossesOnly = {
			type = "toggle",
			name = L["Bosses Only"],
			desc = L["Only record boss fights"],
			order = 100,
		},
		notifications = {
			type = "toggle",
			name = L["Notifications"],
			desc = L["Display start and end statistics of fights. More to remind you that data is being stored and munching your addon ram up than anything else"],
			order = 110,
		},
		minimumDuration = {
			type = "range",
			name = L["Minimum Duration"],
			desc = L["Set the minimum duration for a fight to be stored for (last fight is always stored temporarily)"],
			order = 120,
			min = 0,
			max = 600,
			step = 1,
		},
		keep = {
			type = "range",
			name = L["History Size"],
			desc = L["How many fights to keep"],
			order = 130,
			min = 1,
			max = 500,
			step = 1,
		},
		recordDPS = {
			type = "toggle",
			name = L["Record DPS"],
			desc = L["Records DPS with a fight. Just the basics; damage done to each mob per second by any source in your group. So you can see how raid debuffs are affecting things."],
			order = 140,
		},
		display = {
			type = "group",
			name = L["Display"],
			desc = L["Display settings"],
			guiInline = true,
			get = function(info) return db[info[#info]] end,
			set = function(info, value) db[info[#info]] = value end,
			order = 300,
			args = {
				coloursExtreme = {
					type = "toggle",
					name = L["Coloured Categories"],
					desc = L["Colour the category names to reflect all the class colours applicable (Bring Shades)"],
					order = 10,
		        },
			},
		},
	},
}

local FRIEND_FLAGS = COMBATLOG_OBJECT_AFFILIATION_MINE + COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID
local ENEMY_FLAGS = COMBATLOG_OBJECT_REACTION_HOSTILE + COMBATLOG_OBJECT_REACTION_NEUTRAL
local EXLCUDE_FLAGS = COMBATLOG_OBJECT_TYPE_OBJECT + COMBATLOG_OBJECT_TYPE_PLAYER

-- MakeDestKey
-- Make a compressed Mob GUID for storing, and an ID -> Name lookup attached to the fight data
-- 0xF530000ABCD000123 becomes 43981:1
-- 0xF530000ABCD000492 becomes 43981:2
function module:MakeDestKey(fight, name, guid)
	local mobID = tonumber(guid:sub(-12, -7), 16)
	local mobUnique = tonumber(guid:sub(-6), 16)

	local lookup = self.fight.mobLookup[mobID]
	if (not lookup) then
		lookup = new()
		self.fight.mobLookup[mobID] = lookup
	end
	local key = lookup[mobUnique]
	if (not key) then
		key = (self.lastID[mobID] or 0) + 1
		self.lastID[mobID] = key
		lookup[mobUnique] = key
	end

	local mob = fight.mobs[mobID]
	if (not mob) then
		if (not name) then
			return
		end
		fight.mobs[mobID] = name
	end
	if (key == 1) then
		return mobID
	end
	return format("%d:%d", mobID, key)
end

-- MakeSourceKey
-- Make a compressed Player GUID for storing, and an ID -> Name lookup attached to the fight data
-- 0x0100000000009AFEB becomes 9AFEB
function module:MakeSourceKey(fight, name, guid)
	local playerID = tonumber(guid:sub(-8), 16)
	local player = fight.players[playerID]
	if (not player) then
		fight.players[playerID] = name
	end
	return format("%X", playerID)
end

-- BuildStoredData
function module:BuildStoredData(fight, timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId)
	local playerID = tonumber(sourceGUID:sub(-8), 16)
	local sourceKey = self:MakeSourceKey(fight, sourceName, sourceGUID)
	timeEnd = timeEnd and (timeEnd - timeStart)
	timeStart = max(0, timeStart - fight.combatStart)
	return format("%.3f,%.3f,%s,%d,%s", timeStart, timeEnd or 0, sourceKey, spellId, realSpellId or "")
end

-- GetStoredData
function module:GetStoredData(fight, data)
	local timeStart, timeEnd, sourceID, spellId, realSpellId = strsplit(",", data)
	sourceID = tonumber(sourceID, 16)
	timeStart = tonumber(timeStart) + fight.combatStart
	return timeStart, tonumber(timeEnd) + timeStart, fight.players[sourceID], format("0x010%013X", sourceID), tonumber(spellId), tonumber(realSpellId)
end

-- MobIDFromKey
local function MobIDFromKey(key)
	if (type(key) == "number") then
		return key
	else
		local mobid, index = strsplit(":", key)
		mobid = tonumber(mobid)
		index = index and tonumber(index) or nil
		return mobid, index
	end
end

-- MakeMobGUID
local function MakeMobGUID(key)
	local id, ind = MobIDFromKey(key)
	return format("0xF53%07X%06X", id, ind or 1)
end

-- AddAura
function module:AddAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, realSpellId, realSpellName, auraType)
	if (destName and sourceName and dlookup[spellName]) then

		local target = active[destGUID]
		if (not target) then
			target = new()
			active[destGUID] = target
		end

		local source = target[sourceGUID]
		if (not source) then
			source = new()
			target[sourceGUID] = source
		end

		local s = source[spellName]
		if (not s) then
			local n = new()
			n.timeStart = timestamp
			n.spellId = spellId
			n.realSpellId = realSpellId

			source[spellName] = n
		end

		self.lastEventTime = timestamp
		self.applyCount = self.applyCount + 1
	end

	local mapList = mlookup[spellName]
	if (mapList) then
		for i,map in ipairs(mapList) do
			local mapping = dlookup[map.key]
			if (not mapping or not mapping.requiredTalent or Utopia:HasTalent(sourceName, mapping.requiredTalent)) then
				self:AddAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, mapping and mapping.id or 0, map.key, realSpellId or spellId, realSpellName or spellName, auraType)
			end
		end
	end
end

-- RemoveAura
function module:RemoveAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, realSpellId, realSpellName, auraType)
	if (dlookup[spellName]) then
		local target = active[destGUID]
		if (target) then
			local source = target[sourceGUID]
			if (source) then
				local s = source[spellName]
				if (s) then
					source[spellName] = nil

					if (timestamp > s.timeStart + 0.0009) then	-- Skip things that disappear the same timestamp they apply
						local entry = self:BuildStoredData(self.fight, s.timeStart, timestamp, sourceName, sourceGUID, spellId, realSpellId)
						local destKey = self:MakeDestKey(self.fight, destName, destGUID)
						if (destKey) then
							local data = self.fight.data[destKey]
							if (not data) then
								data = new()
								self.fight.data[destKey] = data
							end
							tinsert(data, entry)
						end
					end

					self.lastEventTime = timestamp
					self.removeCount = self.removeCount + 1
					del(s)
				end
			end
		end
	end

	local mapList = mlookup[spellName]
	if (mapList) then
		for i,map in ipairs(mapList) do
			local mapping = dlookup[map.key]
			if (not mapping or not mapping.requiredTalent or Utopia:HasTalent(sourceName, mapping.requiredTalent)) then
				module:RemoveAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, mapping and mapping.id or 0, map.key, realSpellId or spellId, realSpellName or spellName, auraType)
			end
		end
	end
end

-- RecordDamage
function module:RecordDamage(timestamp, destGUID, destName, amount)
	local f = self.fight

	local key = self:MakeDestKey(f, destName, destGUID)
	if (key) then
		local index = floor(timestamp - f.combatStart) - 1
		local target = self.dps[key]
		if (not target) then
			target = new()
			self.dps[key] = target
		end

		for i = #target + 1,index - 1 do
			tinsert(target, 0)
		end

		target[index] = (target[index] or 0) + amount
	end
end

local clEvents = {}
local dpsEvents = {}

-- SPELL_AURA_APPLIED
function clEvents:SPELL_AURA_APPLIED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags,
	spellId, spellName, spellSchool, auraType)

	if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
		if (band(destFlags, ENEMY_FLAGS) ~= 0) then
			if (band(destFlags, EXLCUDE_FLAGS) == 0) then
				if (not db.bossesOnly or BossIDWeakLookup[destGUID]) then
					self:AddAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, nil, nil, auraType)
				end
			end
		end
	end
end

-- SPELL_AURA_REMOVED
function clEvents:SPELL_AURA_REMOVED(timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags,
	spellId, spellName, spellSchool, auraType)

	if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
		if (band(destFlags, ENEMY_FLAGS) ~= 0) then
			if (band(destFlags, EXLCUDE_FLAGS) == 0) then
				if (not db.bossesOnly or BossIDWeakLookup[destGUID]) then
					self:RemoveAura(timestamp, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, spellId, spellName, nil, nil, auraType)
				end
			end
		end
	end
end

-- SPELL_DAMAGE
function dpsEvents:SWING_DAMAGE(timestamp, destGUID, destName, amount, overkill)
	self:RecordDamage(timestamp, destGUID, destName, amount - overkill)
end

-- SPELL_DAMAGE
function dpsEvents:SPELL_DAMAGE(timestamp, destGUID, destName, spellId, spellName, spellSchool, amount, overkill)
	self:RecordDamage(timestamp, destGUID, destName, amount - overkill)
end
dpsEvents.SPELL_PERIODIC_DAMAGE = dpsEvents.SPELL_DAMAGE

local huntersMark = GetSpellInfo(1130)				-- Hunter's Mark

-- COMBAT_LOG_EVENT_UNFILTERED
function module:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
	if (not active) then
		-- Check for start of a fight by someone in group, first events will happen before you enter combat
		-- and we don't want to lose anything
		if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
			if (band(destFlags, ENEMY_FLAGS) ~= 0) then
				if (band(destFlags, EXLCUDE_FLAGS) == 0) then
					if (UnitInRaid(sourceName) or UnitInParty(sourceName)) then
						if (event == "SPELL_AURA_APPLIED" or event == "SPELL_AURA_REMOVED") then
							local spellId, spellName, spellSchool, auraType = ...
							if (spellName == huntersMark) then
								-- Don't start a fight with Hunter's Mark
								return
							end
						end

						self:StartFighting()
					end
				end
			end
		end
	end
	if (active) then
		local f = clEvents[event]
		if (f) then
			f(self, timestamp, event, sourceGUID, sourceName, sourceFlags, destGUID, destName, destFlags, ...)
		elseif (self.dps) then
			f = dpsEvents[event]
			if (f) then
				if (band(sourceFlags, FRIEND_FLAGS) ~= 0) then
					if (band(destFlags, ENEMY_FLAGS) ~= 0) then
						f(self, timestamp, destGUID, destName, ...)
					end
				end
			end
		end
	end
end

-- PLAYER_REGEN_DISABLED
function module:PLAYER_REGEN_DISABLED()
	if (not Utopia:ShouldShow()) then
		return
	end
	self:StartFighting()
end

-- StartFighting
function module:StartFighting()
	self:EndFighting()

	--d("Fight Started: %s", date("%X",time()))
	self.active = {}
	self.fight = {}
	self.fight.data = {}
	self.fight.players = {}
	self.fight.mobs = {}
	self.fight.combatStart = time()
	self.lastID = {}
	self.fight.mobLookup = {}
	if (db.recordDPS) then
		self.dps = {}
	end

	active = self.active

	self.combatStart = time()
	self.applyCount, self.removeCount = 0, 0

	dlookup = Utopia.lookup.debuffs
	blookup = Utopia.lookup.buffs
	mlookup = Utopia.mappingLookup
end

-- PLAYER_LEAVING_WORLD
function module:PLAYER_LEAVING_WORLD()
	self:CheckEndOfFight()
end

-- PLAYER_REGEN_ENABLED
function module:PLAYER_REGEN_ENABLED()
	if (not self:CheckEndOfFight()) then
		self:RegisterEvent("UNIT_DIED")
		if (not self.fightEndTimer) then
			self.fightEndTimer = self:ScheduleRepeatingTimer("CheckEndOfFight", 2)
		end
	end
end

-- UNIT_DIED
function module:UNIT_DIED()
	self:CheckEndOfFight()
end

-- CheckEndOfFight
function module:CheckEndOfFight()
	if (not InCombatLockdown()) then
		for unit in Utopia:IterateRoster() do
			if (UnitAffectingCombat(unit)) then
				return
			end
		end
		self:EndFighting()
		return true
	end
end

-- EndFighting
function module:EndFighting()
	if (self.fightEndTimer) then
		self:CancelTimer(self.fightEndTimer)
		self.fightEndTimer = nil
	end
	if (not self.fight) then
		return
	end

	--d("Fight Ended: %s", date("%X",time()))
	self:UnregisterEvent("UNIT_DIED")
	self:UnregisterEvent("PLAYER_LEAVING_WORLD")

	if (not db.history) then
		db.history = {}
	end

	self:Cleanup()

	if (time() - self.fight.combatStart > db.minimumDuration) then
		if (next(self.fight.data)) then
			self.lastID = deepDel(self.lastID)
			self.fight.mobLookup = deepDel(self.fight.mobLookup)
			self.fight.combatEnd = self.lastEventTime

			self.fight.player = UnitName("player")
			tinsert(db.history, self.fight)

			if (db.notifications) then
				local link = format("|cFF80FF80|Hutopia_fight:%d|h[%s]|h|r", self.fight.combatStart, L["View"])
				Utopia:Print(format(L["Recorded |cFFFFFF80%s|r (%d secs) - |cFF80FF80%d|r auras %s"], date("%X"), time() - self.fight.combatStart, max(self.applyCount, self.removeCount), link))
			end

			if (self.dps) then
				self.fight.dps = {}
				for mobKey,data in pairs(self.dps) do
					self.fight.dps[mobKey] = table.concat(data, ",")
				end
				self.dps = deepDel(self.dps)
			end

			self.fight = nil
		end
	end

	self.active = deepDel(self.active)
	self.fight = deepDel(self.fight)
	active = nil
end

-- SetItemRef
function module:SetItemRef(link, text, button)
	local cmd, when = strsplit(":", link)
	if (cmd == "utopia_fight") then
		when = tonumber(when)
		for i,fight in ipairs(db.history) do
			if (fight.combatStart == when) then
				module:OpenView(i)
				break
			end
		end
		return true
	end
	return self.hooks.SetItemRef(link, text, button)
end

-- PLAYER_TARGET_CHANGED
function module:PLAYER_TARGET_CHANGED()
	if (UnitExists("target")) then
		local name = UnitName("target")
		local classification = UnitClassification("target")
		self.key = name.."-"..date()
	end
end

-- Cleanup
function module:Cleanup()
	local purged = 0
	for i = #db.history,1,-1 do
		local fight = db.history[i]
		if (not fight.keep and fight.combatEnd - fight.combatStart < db.minimumDuration) then
			purged = purged + 1
			deepDel(tremove(db.history, i))
		end
	end

	for i = #db.history,1,-1 do
		if (#db.history <= db.keep) then
			break
		end

		local fight = db.history[i]
		if (not fight.keep) then
			purged = purged + 1
			deepDel(tremove(db.history, i))
		end
	end
end

-- GetUptime
-- Get's the uptime of a spell, defaulting to the last fight
function module:GetUptime(spellName)
	local data = db.history
	if (data) then
		data = data[#data]
		local spells = data and data.spell and data.spell[spellName]
		if (spells) then
			local linear = data and data.linear
			if (linear) then
				local startTime = data.combatStart
				local endTime = data.combatEnd
				local on, off = 0, 0
				for i,info in ipairs(spells) do
					on = on + (info.timeStart - startTime)
					if (info.timeEnd) then
						off = off + (info.timeEnd - startTime)
					else
						off = off + (endTime - startTime)
					end
				end

				on = on / #spells
				off = off / #spells

				local buffTime = off - on
				local fightTime = endTime - startTime

				return buffTime, fightTime
			end
		end
	end
end

local bgDef = {edgeInsets = {left = 6, right = 6, top = 6, bottom = 6}, edgeSize = 10, tileSize = 32, edgeFile = "Interface\\Tooltips\\UI-Tooltip-Border"}
function module:SetBackdrop(frame, back)
	bgDef.bgFile = back and "Interface\\Tooltips\\UI-Tooltip-Background" or nil
	frame:SetBackdrop(bgDef)
	frame:SetBackdropBorderColor(1, 1, 1, 0.8)
	if (back) then
		frame:SetBackdropColor(0, 0, 0, 1)
	end
end	

-- OpenView
function module:OpenView(key)
	if (not self.freeFrames) then
		self.freeFrames = {}
		self.frames = {}
	end
	local frame = tremove(self.freeFrames) or self:CreateUpTimeFrame()
	self.frames[frame] = true

	frame:SetView(key)
	return frame
end

do
	-- OnHide
	local function frameOnHide(self)
		self.offset = nil
		self.fight = nil
		self.area.fight = nil
		self.area:OnClose()
		self:ClearSelected()
		if (self.dpsFrame) then
			self.showDPS = nil
			self:ShowDPS()
		end

		module.frames[self] = nil
		tinsert(module.freeFrames, self)

		self.orderedMobNames = del(self.orderedMobNames)
	end

	-- SetTitle
	local function frameSetTitle(self)
		local when = module:DescribeDate(self.fight.combatStart)
		local dur = self.fight.combatEnd - self.fight.combatStart
		if (dur > 60) then
			dur = format(MINUTES_ABBR.." "..SECONDS_ABBR, dur / 60, dur % 60)
		else
			dur = format(SECONDS_ABBR, dur)
		end
		self.title:SetFormattedText("%s: %s %s - %s [%s]", Utopia:ColourPlayer(self.fight.player), when, date("%X", self.fight.combatStart), date("%X", self.fight.combatEnd), dur)
	end

	-- OnShow
	local function frameOnShow(self)
		self:SetTitle()
		self:ClearSelected()
		self:OrderMobNames()
		self:PopulateMobNames()
		self.list:Select(1)
	end

	local function framelistOnVerticalScroll(self, offset)
		self:GetParent().offset = offset
		self:GetParent():PopulateMobNames()
	end

	local function framelistOnClick(self, ...)
		local frame = self:GetParent():GetParent()
		frame:ClearSelected()

		local count = 0
		for i = 1,select("#",...) do
			local index = select(i, ...)
			frame:AddSelected(index)
			count = count + 1
		end

		frame.area:SelectMobKey()
	end

	local function frameClearSelected(self)
		del(self.area.mobKeys)
		self.area.mobKeys = new()
	end

	local function frameAddSelected(self, index)
		local name = self.orderedMobNames[index]
		self.area.mobKeys[self.mobnameToKeyLookup[name]] = true
	end

	-- SetView
	local function frameSetView(self, key)
		self.key = key
		local fight = db.history[key]
		assert(fight)
		self.fight = fight
		self.area.fight = fight

		self:Hide()
		self:Show()
	end

	-- PopulateMobNames
	local function framePopulateMobNames(self)
		self.list:Select()

		local lines = floor(self.list:GetHeight() / 16)
		for i = 1,lines do
			local name = self.orderedMobNames[i + self.offset]
			if (not name) then
				break
			end

			local b = self.list.lines[i]
			b.key = self.mobnameToKeyLookup[name]
			b:SetText(name)
			b:Show()
		end
		self.list.scroll:Update(#self.orderedMobNames)

		for i = #self.orderedMobNames + 1,#self.list.lines do
			self.list.lines[i]:Hide()
		end
	end

	-- OrderMobNames
	local function frameOrderMobNames(self)
		self.offset = 0
		self.orderedMobNames = del(self.orderedMobNames)
		self.mobnameToKeyLookup = del(self.mobnameToKeyLookup)
		self.orderedMobNames = new()
		self.mobnameToKeyLookup = new()
		local l = self.mobnameToKeyLookup
		local fight = self.fight
		local mobs = self.fight.mobs

		for mobKey,info in pairs(fight.data) do
			local mobid, index = MobIDFromKey(mobKey)
			local baseName = mobs[mobid]
			local name

			if (index and index > 1) then
				name = format("%s (%d)", baseName, index)
			else
				name = baseName
			end
			if (l[name]) then
				-- Two mobs of same name can have a different ID, so we run into the problem of duplicates here
				-- ie: Tempest Minion with Emalon; the spawned ones during fight have a different ID than the 4
				-- static ones at start
				for i = 2,10000 do
					local temp = format("%s (%d)", baseName, i)
					if (not l[temp]) then
						name = temp
						break
					end
				end
			end

			tinsert(self.orderedMobNames, name)
			l[name] = mobKey
		end
		sort(self.orderedMobNames)
	end

	local function setModeCat(self)
		self:GetParent().area:SetMode("category")
	end
	local function setModeSpell(self)
		self:GetParent().area:SetMode("spell")
	end

	local function getPlayerNames(self)
		return self:GetParent().area.playerNames
	end

	local function setPlayerNames(self, newval)
		self:GetParent().area.playerNames = newval
		self:GetParent().area:SelectMobKey()
	end

	local function getShowDPS(self)
		local frame = self:GetParent()
		self:SetEnabled(frame.fight.dps)
		return frame.showDPS
	end

	local function setShowDPS(self, newval)
		self:GetParent().showDPS = newval
		self:GetParent():ShowDPS()
	end

	local function frameGetKeep(self)
		local frame = self:GetParent()
		return frame.fight and frame.fight.keep
	end

	local function frameSetKeep(self, newval)
		local frame = self:GetParent()
		frame.fight.keep = newval and true
	end

	-- frame.ShowDPS
	local function frameShowDPS(self)
		if (self.showDPS) then
			if (not self.dpsFrame) then
				self.dpsFrame = module:CreateDPSFrame(self)
				self.horizontalBar = module:CreateHorizontalBar(self.dpsFrame, "TOP")
				self.horizontalBar:SetPoint("TOPRIGHT", self.dpsFrame, "TOPRIGHT", 24, 8)

				self.dpsFrame:SetPoint("BOTTOMRIGHT", -40, 43)
			else
				self.dpsFrame:Show()
			end
			self.dpsFrame:SetPoint("TOPLEFT", self.list, "BOTTOMRIGHT", 10, self.dpsFrame.height)
			self.dpsFrame:SetFight(self.fight)

			self.area:SetPoint("BOTTOMRIGHT", self.dpsFrame, "TOPRIGHT", 0, 6)
		else
			if (self.dpsFrame) then
				self.dpsFrame:Hide()
				self.dpsFrame:SetPoint("TOPLEFT", self.list, "BOTTOMRIGHT", 10, 0)
				self.area:SetPoint("BOTTOMRIGHT", self.dpsFrame, "TOPRIGHT", 0, 0)
			end
		end
	end

	-- CreateUpTimeFrame
	function module:CreateUpTimeFrame()
		self.frameNo = (self.frameNo or 0) + 1
		local frame = CreateFrame("Frame", "UtopiaUpTimeFrame"..self.frameNo, UIParent)
		frame:Hide()
		frame:SetScale(0.8)
		frame:SetWidth(768)
		frame:SetHeight(512)
		self:ApplyBackground(frame)
		self:MovableFrame(frame)

		frame.list = self:CreateListFrame(frame, true, true)
		frame.list:SetPoint("TOPLEFT", 24, -82)
		frame.list:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", 195, 42)
		frame.list.OnClick = framelistOnClick
		frame.list.OnVerticalScroll = framelistOnVerticalScroll

		frame.bar = self:CreateVerticalBar(frame.list, "RIGHT", true)

		frame.button1 = self:Button(frame, L["By Category"], setModeCat, "BOTTOMLEFT", 15, 15)
		frame.button2 = self:Button(frame, L["By Spell"], setModeSpell, "LEFT", 10, 0, frame.button1, "RIGHT")
		frame.checkbox1 = self:CheckBox(frame, L["Player Names"], getPlayerNames, setPlayerNames, "LEFT", 10, 0, frame.button2, "RIGHT")
		frame.dpsCheckbox = self:CheckBox(frame, L["Show DPS"], getShowDPS, setShowDPS, "LEFT", 100, 0, frame.checkbox1, "RIGHT")

		local area = self:CreateWorkArea(frame)
		frame.area = area
		area:SetPoint("TOPLEFT", frame.list, "TOPRIGHT", 10, 0)
		area:SetPoint("BOTTOMRIGHT", -40, 43)

		-- Keep option
		frame.keep = module:CheckBox(frame, L["Keep"], frameGetKeep, frameSetKeep, "TOPLEFT", 75, -35)

		frame.OnShow			= frameOnShow
		frame.OnHide			= frameOnHide
		frame.SetTitle			= frameSetTitle
		frame.OrderMobNames		= frameOrderMobNames
		frame.PopulateMobNames	= framePopulateMobNames
		frame.SetView			= frameSetView
		frame.ClearSelected		= frameClearSelected
		frame.AddSelected		= frameAddSelected
		frame.ShowDPS			= frameShowDPS

		frame:SetScript("OnShow", frame.OnShow)
		frame:SetScript("OnHide", frame.OnHide)

		return frame
	end
end

do
	local areaCreateSpellLine

	-- Init
	local function areaInit(self)
		self:ClearDisplay()
		self.nameToLineLookup = new()
		self.dataIndex = new()
		self:BuildSpellNameList()
	end

	local function areaSetMode(self, mode)
		self.mode = mode
		self:Init()
		self:Populate()
	end

	local function areaSelectMobKey(self)
		self:Init()
		self:Populate()
	end

	-- OnClose
	local function areaOnClose(self)
		self.scale = 6
		self.playerNames = nil
		self.lastTooltipInfo = del(self.lastTooltipInfo)
		self:ClearDisplay()
		self.rightScroll:SetHorizontalScroll(0)
		self.rightScroll:SetVerticalScroll(0)
		self.leftScroll:SetVerticalScroll(0)
	end

	-- ClearDisplay
	local function areaClearDisplay(self)
		self.offset = nil
		self.dataIndex = del(self.dataIndex)
		self.nameToLineLookup = deepDel(self.nameToLineLookup)
		for i,line in ipairs(self.lines) do
			line.info = nil
			line.spellId = nil
			line:UnlockHighlight()
			line:Hide()
		end
		self.offset = 0
		self.shownLines = 0
		self.rowHighlight = nil
	end

	-- RightScrollOnUpdate
	local function areaRightScrollOnUpdate(self, elapsed)
		self:GetParent():ShowCursorLocation()
	end

	-- RightScrollOnEnter
	local function areaRightScrollOnEnter(self)
		if (not self.dragging and not self.speedX) then
			self:SetScript("OnUpdate", areaRightScrollOnUpdate)
		end
	end

	-- RightScrollOnLeave
	local function areaRightScrollOnLeave(self)
		local area = self:GetParent()
		if (not self.dragging and not self.speedX) then
			self:SetScript("OnUpdate", nil)
			area:HideCursorLocation()
		end
		area:RemoveHighlight()
		GameTooltip:Hide()
	end

	-- rightscrollOnUpdateCheckScrollRange
	local function rightscrollOnUpdateCheckScrollRange(self)
		self.onUpdateCheck = self.onUpdateCheck + 1
		if (self.onUpdateCheck > 1) then
			local offset = -self:GetHorizontalScroll()
			local newoffset = max(0, min(self:GetHorizontalScrollRange(), offset))
			if (offset ~= newoffset) then
				self:SetHorizontalScroll(-newoffset)
			end
			if (MouseIsOver(self)) then
				self:SetScript("OnUpdate", areaRightScrollOnUpdate)
			else
				self:SetScript("OnUpdate", nil)
			end

			local area = self:GetParent()
			area.lastX, area.lastY = nil, nil
			area:ShowCursorLocation()
		end
	end

	-- Populate
	local function areaPopulate(self)
		if (not self.mobKeys) then
			return
		end

		local fight = self.fight
		local data = fight.data

		local w = max(200, (self.fight.combatEnd - self.fight.combatStart) * self.scale)
		self.rightScrollChild:SetWidth(w)

		local parent = self:GetParent()
		local keyLookup = parent.mobnameToKeyLookup

		for i,mobName in ipairs(parent.orderedMobNames) do
			local mobKey = keyLookup[mobName]

			if (self.mobKeys[mobKey]) then
				local fightData = data[mobKey]
				local index = self.dataIndex[mobKey] or 1

				for i = index,#fightData do
					local segment = fightData[i]
					local timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId = module:GetStoredData(fight, segment)
					if (timeEnd > timeStart + 0.0009) then		-- Sanity
						self:AddSpell(timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId, mobName, mobKey)
					end
				end

				self.dataIndex[mobKey] = #fightData + 1
			end
		end

		self.rightScroll.onUpdateCheck = 0
		self.rightScroll:SetScript("OnUpdate", rightscrollOnUpdateCheckScrollRange)

		self:ShowScroll()
		self:ShowScale()

		if (parent.showDPS and parent.dpsFrame) then
			parent.dpsFrame:SetFight(self.fight)
		end
	end

	local function areaCreateSlidingIndicator(self)
		local pointer = self:CreateTexture(nil, "OVERLAY")
		pointer:SetTexture("Interface\\Addons\\Utopia\\Textures\\SlidingIndicator")
		pointer:SetWidth(24)
		pointer:SetHeight(8)
		pointer:SetTexCoord(0, 1, 0.05, 0.5)

		local text = self:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		text:SetPoint("BOTTOM", pointer, "TOP", 0, -3)

		pointer.text = text
		return pointer
	end

	-- HideCursorLocation
	local function areaHideCursorLocation(self)
		if (self.cursorTick) then
			self.cursorTick:Hide()
			self.cursorTick.text:Hide()
		end
	end

	-- ShowCursorLocation
	local function areaShowCursorLocation(self)
		if (not MouseIsOver(self)) then
			self:HideCursorLocation()
			self:RemoveHighlight()
			GameTooltip:Hide()
			self.lastTooltipInfo = del(self.lastTooltipInfo)
			return
		end

		local x, y = GetCursorPosition()
		x, y = x / self:GetEffectiveScale(), y / self:GetEffectiveScale()
		if (self.lastX == x and self.lastY == y) then
			return
		end
		self.lastX, self.lastY = x, y

		local offset = -self.rightScroll:GetHorizontalScroll()
		local left = self.rightScroll:GetLeft()
		local cursorTime = ((x - left) + offset) / self.scale
		self.cursorTime = cursorTime

		local startSec = offset / self.scale
		local endSec = (offset + self.rightScroll:GetWidth() - 1) / self.scale

		local tooltipShown
		if (cursorTime < startSec or cursorTime > endSec) then
			self:HideCursorLocation()
			self:RemoveHighlight()
		else
			if (not self.cursorTick) then
				self.cursorTick = self:CreateSlidingIndicator()
				self.cursorTick:SetHeight(24)
				self.cursorTick:SetVertexColor(0, 1, 0)
				self.cursorTick:SetTexCoord(0, 1, 0.14, 0.9)
			end

			self.cursorTick:Show()
			self.cursorTick:SetPoint("BOTTOM", self.timeStart, "BOTTOM", (cursorTime - startSec) * self.scale, 0)

			if (cursorTime < 60) then
				self.cursorTick.text:SetFormattedText("%.3fs", cursorTime)
			else
				self.cursorTick.text:SetFormattedText("%dm %.3fs", cursorTime / 60, cursorTime % 60)
			end
			self.cursorTick.text:Show()

			-- Now highlight the spell row for where we are
			local left, top, width, height = self.lines[1]:GetRect()
			local row = floor((top - y) / height) + 2

			if (self.rowHighlight ~= row) then
				self:RemoveHighlight()
			end
			if (row >= 1 and row <= #self.lines) then
				local line = self.lines[row]
				line:LockHighlight()
				self.rowHighlight = row

				-- Now see if any spells are under the cursor, and show the info
				-- We cache what we show and compare this with last tooltip. Because we're updating
				-- so often there's a noticable lag when updating the tooltip at this frequency
				local temp = new()
				local cursorTimestamp = self.fight.combatStart + cursorTime
				for i,block in ipairs(line.blocks.list) do
					if (cursorTimestamp >= block.timeStart and cursorTimestamp <= block.timeEnd) then
						local name, rank, tex = GetSpellInfo(block.spellId)
						local r, g, b = block:GetVertexColor()
						tinsert(temp, new(name, tex, block.sourceName, r, g, b, block.mobName))
					end
				end

				if (#temp > 0) then
					local redraw
					if (self.lastTooltipInfo) then
						if (#self.lastTooltipInfo ~= #temp) then
							redraw = true
						else
							for i = 1,#temp do
								local t1 = temp[i]
								local t2 = self.lastTooltipInfo[i]
								if (t1[1] ~= t2[1] or t1[3] ~= t2[3]) then
									redraw = true
									break
								end
							end
						end
					else
						redraw = true
					end

					local showTime = self.fight.combatStart + cursorTime
					showTime = format("%s.%03d", date("%X", showTime), showTime * 1000 % 1000)

					tooltipShown = true
					if (redraw) then
						local c = 0
						for key in pairs(self.mobKeys) do
							c = c + 1
						end

						GameTooltip:SetOwner(line, "ANCHOR_CURSOR")
						GameTooltip:SetText(showTime, 1, 1, 1)

						for i,t in ipairs(temp) do
							if (c > 1) then
								GameTooltip:AddDoubleLine(format("|cFFFFFF80%s|r on %s", t[1], t[7]), t[3], 0.5, 0.5, 0.5, t[4], t[5], t[6])
							else
								GameTooltip:AddDoubleLine(t[1], t[3], 1, 1, 0.5, t[4], t[5], t[6])
							end
							GameTooltip:AddTexture(t[2])
						end
						GameTooltip:Show()
					else
						-- Always update the cursor location time
						GameTooltipTextLeft1:SetText(showTime)
						GameTooltipTextLeft1:SetTextColor(1, 1, 1)
					end
				end

				del(self.lastTooltipInfo)
				self.lastTooltipInfo = temp
			end
		end

		if (not tooltipShown) then
			GameTooltip:Hide()
			self.lastTooltipInfo = del(self.lastTooltipInfo)
		end
	end

	-- RemoveHighlight
	local function areaRemoveHighlight(self)
		if (self.rowHighlight) then
			self.lines[self.rowHighlight]:UnlockHighlight()
			self.rowHighlight = nil
		end
	end

	-- ShowScroll
	local function areaShowScroll(self)
		local offset = -self.rightScroll:GetHorizontalScroll()

		if (not self.timeStart) then
			self.timeStart = self:CreateSlidingIndicator()
			self.timeStart:SetPoint("BOTTOM", self.rightScroll, "TOPLEFT")
		end
		if (not self.timeEnd) then
			self.timeEnd = self:CreateSlidingIndicator()
			self.timeEnd:SetPoint("BOTTOM", self.rightScroll, "TOPRIGHT", -1, 0)
		end

		local width = self.rightScroll:GetWidth() - 1
		local startSec = offset / self.scale
		local endSec = (offset + width) / self.scale

		self.timeStart.text:SetFormattedText("%d", abs(startSec))
		self.timeEnd.text:SetFormattedText("%d", endSec)

		if (not self.ticks) then
			self.ticks = {}
		end

		local div = self.scale < 4 and 60 or 15

		local startPoint = floor((startSec + div) / div) * div
		local i = 0
		for secs = startPoint,endSec,div do
			local short = secs % 30 ~= 0
			local gap = short and 20 or 50

			local showOffset = (secs - startSec) * self.scale

			if (showOffset >= width - 10) then
				break
			elseif (showOffset > 10) then
				local nearEdge = min(showOffset - 10, width - 10 - showOffset)
				local fade
				if (nearEdge < gap) then
					fade = nearEdge / gap
				else
					fade = 1
				end

				i = i + 1
				local tick = self.ticks[i]
				if (not tick) then
					tick = self:CreateSlidingIndicator()
					self.ticks[i] = tick
				end

				tick:SetPoint("BOTTOM", self.timeStart, "BOTTOM", (secs - startSec) * self.scale, 0)
				tick:Show()

				if (short) then
					tick:SetHeight(3)
					tick:SetTexCoord(0, 1, 0.5, 0.9)
					tick.text:Hide()
				else
					tick:SetHeight(8)
					tick:SetTexCoord(0, 1, 0.05, 0.5)
					tick.text:SetFormattedText("%d:%02d", secs / 60, secs % 60)

					tick.text:Show()
				end

				tick:SetAlpha(fade)
				tick.text:SetAlpha(fade)
			end
		end
		for j = i + 1,#self.ticks do
			local tick = self.ticks[j]
			tick:Hide()
			tick.text:Hide()
		end
	end

	-- BuildSpellNameList
	local function areaBuildSpellNameList(self)
		if (not self.mobKeys) then
			return
		end

		local fight = self.fight
		local data = fight.data

		local spellIcons = new()
		local spellIds = new()
		local lineColours = new()
		local activeClassesForCat = new()

		local temp = new()
		for mobKey in pairs(self.mobKeys) do
			local fightData = data[mobKey]

			for i = 1,#fightData do
				local segment = fightData[i]
				local timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId = module:GetStoredData(fight, segment)
				local name, rank, tex = GetSpellInfo(spellId)
				local spellInfo = Utopia.lookup.debuffs[name]
				if (self.mode == "category") then
					if (spellInfo and db.coloursExtreme) then
						name = spellInfo.key
						local cat = Utopia.debuffs[name]

						local temp2 = activeClassesForCat[cat.key]
						if (not temp2) then
							temp2 = new()
							activeClassesForCat[cat.key] = temp2
						end
						temp2[spellInfo.class] = true
					end
				else
					spellIcons[name] = tex
					spellIds[name] = spellId
					local c = RAID_CLASS_COLORS[spellInfo.class]
					lineColours[name] = new(c.r, c.g, c.b)
				end

				temp[name] = true
			end
		end

		local toSort = new()
		for name in pairs(temp) do
			tinsert(toSort, name)
		end
		del(temp)
		sort(toSort)

		for i,name in ipairs(toSort) do
			local displayLine = self.nameToLineLookup[name]
			if (not displayLine) then
				self.shownLines = self.shownLines + 1
				displayLine = self.shownLines
				self.nameToLineLookup[name] = displayLine
			end

			-- Pre-create the known spell lines, so they appear in correct order later
			local line = self.lines[displayLine]
			if (not line) then
				line = self:CreateSpellLine(displayLine)
			end
			line.text:SetText(name)
			line:Show()

			if (self.mode == "category") then
				local cat = Utopia.debuffs[name]
				line.icon:SetTexture(cat.icon)
				line.text:SetTextColor(0.8, 0.8, 0.8)
			else
				line.icon:SetTexture(spellIcons[name])
			end
			line.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)

			if (self.mode == "spell") then
				line.spellId = spellIds[name]
				line.text:SetTextColor(unpack(lineColours[name]))
			else
				if (db.coloursExtreme) then
					local cat = Utopia.debuffs[name]

					local activeClasses = activeClassesForCat[cat.key]
					local temp = new()
					for class in pairs(activeClasses) do
						tinsert(temp, class)
					end
					sort(temp)

					lineColours[name] = new()
					for i,class in pairs(temp) do
						local c = RAID_CLASS_COLORS[class]
						tinsert(lineColours[name], new(c.r, c.g, c.b))
					end
					del(temp)

					local str
					local lc = lineColours[name]
					if (#lc > 2) then
						str = ""
						for i = 1,#lc - 1 do
							local r1, g1, b1 = unpack(lc[i])
							local r2, g2, b2 = unpack(lc[i + 1])
	
							local part = name:sub(floor((i - 1) * (#name / (#lc - 1)) + 1), floor(i * (#name / (#lc - 1))))
							part = Gradient(part, r1, g1, b1, r2, g2, b2)
							str = str.. part
						end
					elseif (#lc == 2) then
						local r1, g1, b1 = unpack(lc[1])
						local r2, g2, b2 = unpack(lc[2])
						str = Gradient(name, r1, g1, b1, r2, g2, b2)
					else
						line.text:SetTextColor(unpack(lc[1]))
					end
					if (str) then
						line.text:SetText(str)
					end
				end
			end
		end

		--self.leftScrollChild:SetHeight(self.shownLines * 20)
		--self.rightScrollChild:SetHeight(self.shownLines * 20)

		del(toSort)
		del(spellIcons)
		del(spellIds)
		del(lineColours)
	end

	-- AddSpell
	local function areaAddSpell(self, timeStart, timeEnd, sourceName, sourceGUID, spellId, realSpellId, mobName, mobKey)
		local xOffsetStart = (timeStart - self.fight.combatStart) * self.scale
		local xWidth = (timeEnd - timeStart) * self.scale
		if (xWidth < 0.05) then
			return
		end

		local spellName, rank, tex = GetSpellInfo(spellId)
		local spellInfo = Utopia.lookup.debuffs[spellName]
		local name
		if (self.mode == "category") then
			name = spellInfo.key
		else
			name = spellName
		end

		local displayLine = self.nameToLineLookup[name]
		if (not displayLine) then
			self.shownLines = self.shownLines + 1
			displayLine = self.shownLines
			self.nameToLineLookup[name] = displayLine
		end

		local line = self.lines[displayLine]
		if (not line) then
			line = self:CreateSpellLine(displayLine)
		end

		local tex = module:AquireAreaTexture(line.blocks, self.playerNames)
		tex:SetPoint("TOPLEFT", xOffsetStart, 0)
		tex:SetWidth(xWidth)

		local c = RAID_CLASS_COLORS[spellInfo.class]
		local r, g, b = c.r, c.g, c.b
		if (displayLine % 2 == 1) then
			r, g, b = max(0, r * 0.9), max(0, g * 0.9), max(0, b * 0.9)
		end
		tex:SetVertexColor(r, g, b)

		if (self.playerNames) then
			tex:SetOwnerText(sourceName, r, g, b)
		end

		tex.spellId		= spellId
		tex.realSpellId	= realSpellId
		tex.sourceName	= sourceName
		tex.timeStart	= timeStart
		tex.timeEnd		= timeEnd
		tex.mobName		= mobName
		tex.mobKey		= mobKey

		tinsert(line.blocks.list, tex)
	end

	do
		local function slOnMouseDown(self)
			local area = self:GetParent():GetParent():GetParent()
			area.rightScroll:OnMouseDown()
		end

		local function slOnMouseUp(self)
			local area = self:GetParent():GetParent():GetParent()
			area.rightScroll:OnMouseUp()
		end

		local function slOnEnter(self)
			if (self.spellId) then
				local area = self:GetParent():GetParent():GetParent()
				if (not area.rightScroll.dragging) then
					GameTooltip:SetOwner(self, "ANCHOR_TOP")
					GameTooltip:SetHyperlink("spell:"..self.spellId)
				end
			end
		end

		local function slOnLeave(self)
			GameTooltip:Hide()
		end

		local function slOnShow(self)
			-- Also, we need to hide this when we hide the line else the ScrollFrame
			-- will keep the max growth height we've had even with hidden lines
			self.blocks:Show()
		end

		local function slOnHide(self)
			local list = self.blocks.list
			for i,tex in pairs(list) do
				module:ReleaseAreaTexture(tex)
			end
			del(list)
			self.blocks.list = new()
			self.blocks:Hide()
		end

		-- CreateSpellLine
		function areaCreateSpellLine(self, index)
			local l = CreateFrame("Button", nil, self.leftScrollChild)
			self.lines[index] = l
			l:EnableMouse(true)

			-- NOTE: Have to use absolute anchoring here only to the ScrollChild because of ScrollFrame limitations
			l:SetPoint("TOPLEFT", 0, -(index - 1) * 20)
			l:SetPoint("BOTTOMRIGHT", self.leftScrollChild, "TOPRIGHT", 0, -index * 20)

			if (index % 2 == 0) then
				l.bg = l:CreateTexture(nil, "BACKGROUND")
				l.bg:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
				l.bg:SetVertexColor(0.3, 0.3, 0.8, 0.3)
				l.bg:SetAllPoints()
			end

			l.icon = l:CreateTexture(nil, "OVERLAY")
			l.icon:SetWidth(18)
			l.icon:SetHeight(18)
			l.icon:SetPoint("TOPLEFT", 1, -1)

			l.text = l:CreateFontString(nil, "OVERLAY", "GameFontNormal")
			l.text:SetHeight(20)
			l.text:SetWidth(300)
			l.text:SetPoint("LEFT", l.icon, "RIGHT", 2, 0)
			l.text:SetJustifyH("LEFT")

			l:SetHighlightTexture("Interface\\Buttons\\UI-Common-MouseHilight")
			l.highlight = l:GetHighlightTexture()
			l.highlight:SetBlendMode("ADD")
			l.highlight:SetTexCoord(0.25, 0.75, 0.25, 0.75)
			l.highlight:SetAlpha(0.3)
			l.highlight:SetAllPoints(true)

			l.blocks = CreateFrame("Frame", nil, self.rightScrollChild)
			l.blocks.list = {}

			-- NOTE: Have to use absolute anchoring here only to the ScrollChild because of ScrollFrame limitations
			l.blocks:SetPoint("TOPLEFT", 0, -(index - 1) * 20)
			l.blocks:SetPoint("BOTTOMRIGHT", self.rightScrollChild, "TOPRIGHT", 0, -index * 20)

			l:SetScript("OnMouseDown", slOnMouseDown)
			l:SetScript("OnMouseUp", slOnMouseUp)
			l:SetScript("OnEnter", slOnEnter)
			l:SetScript("OnLeave", slOnLeave)
			l:SetScript("OnHide", slOnHide)
			l:SetScript("OnShow", slOnShow)

			return l		
		end
	end


	local areaOnMouseDown, areaOnMouseUp
	do
		-- Dragging Functions
		local function draggingOnUpdate(self, elapsed)
			local x, y = GetCursorPosition()
			x, y = x / self:GetEffectiveScale(), y / self:GetEffectiveScale()

			local curX, curY = -self:GetHorizontalScroll(), self:GetVerticalScroll()
	    	local newX = min(self:GetHorizontalScrollRange(), max(0, curX - x + self.dragX))
	    	local newY = min(self:GetVerticalScrollRange(), max(0, curY + y - self.dragY))

			self:SetHorizontalScroll(-newX)
			self:SetVerticalScroll(newY)
			self:GetParent():ShowCursorLocation()

			local scale = self:GetParent().scale
			self.speedX = (newX - curX)
			self.speedY = newY - curY
			self.dragX, self.dragY = x, y
		end

		-- driftingOnUpdate
		local function driftingOnUpdate(self, elapsed)
			local curX, curY = -self:GetHorizontalScroll(), self:GetVerticalScroll()

			local scale = self:GetParent().scale
			self.speedX = self.speedX * (1 - elapsed * 10)
			self.speedY = self.speedY * (1 - elapsed * 10)

	    	local newX = min(self:GetHorizontalScrollRange(), max(0, curX + self.speedX))
	    	local newY = min(self:GetVerticalScrollRange(), max(0, curY + self.speedY))

			self:SetHorizontalScroll(-newX)
			self:SetVerticalScroll(newY)
			self:GetParent():ShowCursorLocation()

			if (abs(self.speedX) < 0.1 and abs(self.speedY) < 0.1) then
				self.speedX, self.speedY = nil, nil
				if (MouseIsOver(self)) then
					self:SetScript("OnUpdate", areaRightScrollOnUpdate)
				else
					self:SetScript("OnUpdate", nil)
				end
				return
			end
		end

		-- OnMouseDown
		function areaOnMouseDown(self)
			self.dragging = true
			self.dragX, self.dragY = GetCursorPosition()
			self.dragX, self.dragY = self.dragX / self:GetEffectiveScale(), self.dragY / self:GetEffectiveScale()
			self:SetScript("OnUpdate", draggingOnUpdate)
		end

		-- OnMouseUp
		function areaOnMouseUp(self)
			self.dragging = nil
			if (max(abs(self.speedX or 0), abs(self.speedY or 0)) >= 0.1) then
				self:SetScript("OnUpdate", driftingOnUpdate)
			else
				self.speedX, self.speedY = nil, nil
				if (MouseIsOver(self)) then
					self:SetScript("OnUpdate", areaRightScrollOnUpdate)
				else
					self:SetScript("OnUpdate", nil)
				end
			end
		end
	end

	-- Zoom
	local function areaZoom(self, zoomIn)
		-- Work out where our cursor is atm, we'll want it in the same position after zooming
		local offset = -self.rightScroll:GetHorizontalScroll()
		local x = GetCursorPosition() / self:GetEffectiveScale()
		local cursorOffset = x - self.rightScroll:GetLeft()
		local cursorTime = (cursorOffset + offset) / self.scale
		local startSec = offset / self.scale
		if (cursorTime < startSec) then
			cursorTime = startSec	-- Mouse is not over the ScrollFrame (probably using the controls for zooming)
		end

		-- Change the scale
		local diff = floor((self.scale + 10) / 10)			-- 1-10 steps 1, 11-20 steps 2 etc.
		if (zoomIn) then
			self.scale = min(64, self.scale + diff)
		else
			self.scale = max(1, self.scale - diff)
		end

		-- Redraw
		self:Init()
		self:Populate()

		-- Now set the new offset so the cursor is where it was before zooming
		local newOffset = (cursorTime * self.scale) - cursorOffset
		if (newOffset < 0) then
			newOffset = 0
		end
		self.rightScroll:SetHorizontalScroll(-newOffset)
	end

	-- OnMouseWheel
	local function areaOnMouseWheel(self, delta)
		local area = self:GetParent()
		if (delta < 0) then
			area:Zoom()
		else
			area:Zoom(true)
		end
	end

	-- ShowScale
	-- Internally, a scale of 1 means 1 pixel to 1 second
	-- That's a bit small tho, 6 pixels to 1 second is a good visual point, so we'll call that the 100% mark
	local function areaShowScale(self)
		self.scaleFrame.text:SetFormattedText(L["Zoom: %d%%"], 100 / 6 * self.scale)
	end

	-- ScaleUpOnClick
	local function areaScaleUpOnClick(self)
		local area = self:GetParent():GetParent().area
		area:Zoom(true)
	end

	-- ScaleDownOnClick
	local function areaScaleDownOnClick(self)
		local area = self:GetParent():GetParent().area
		area:Zoom()
	end

	-- area.rightScroll:OnSizeChanged
	local function areaOnSizeChanged(self, x, y)
		self:GetParent():ShowScroll()
	end

	-- area:OnVerticalScroll
	local function areaOnVerticalScroll(self, offset)
		local rightScrollbar = getglobal(self:GetName().."ScrollBar")
		self:GetParent().leftScroll:SetVerticalScroll(rightScrollbar:GetValue())
		self:GetParent():ShowScroll()
	end

	-- area.rightScroll:OnHorizontalScroll
	local function areaOnHorizontalScroll(self, offset)
		local scrollbar = getglobal(self:GetName().."ScrollBar2")
		if (scrollbar) then
			scrollbar:SetValue(offset)
			local min, max = scrollbar:GetMinMaxValues()
			if ( offset <= 0 ) then
				getglobal(scrollbar:GetName().."ScrollUpButton"):Disable()
			else
				getglobal(scrollbar:GetName().."ScrollUpButton"):Enable()
			end
			if (offset >= max) then
				getglobal(scrollbar:GetName().."ScrollDownButton"):Disable()
			else
				getglobal(scrollbar:GetName().."ScrollDownButton"):Enable()
			end
		end

		local area = self:GetParent()
		local frame = area:GetParent()
		area:ShowScroll()
		if (frame.dpsFrame and frame.showDPS) then
			frame.dpsFrame.scroll:SetHorizontalScroll(self:GetHorizontalScroll())
		end
	end

	-- CreateWorkArea
	function module:CreateWorkArea(parent)
		local a = CreateFrame("Frame", nil, parent)
		-- info.record["0xF13000808A015559"]["0x010000000008BC20"]["Improved Scorch"]
		a.lines = {}
		a.mode = "category"
		a.scale = 6

		local leftScrollName = self:GenericSuffixName(parent, "ScrollFrame")
		local rightScrollName = self:GenericSuffixName(parent, "ScrollFrame")

		a.leftScroll = CreateFrame("ScrollFrame", leftScrollName, a)
		local ls = a.leftScroll
		a.leftScrollChild = CreateFrame("Frame", nil, ls)

		a.bar = self:CreateVerticalBar(a.leftScroll, "RIGHT", true)

		a.rightScroll = CreateFrame("ScrollFrame", rightScrollName, a, "UIPanelScrollFrameTemplate")
		local rs = a.rightScroll
		a.rightScrollChild = CreateFrame("Frame", nil, rs)

		ls:SetScrollChild(a.leftScrollChild)
		rs:SetScrollChild(a.rightScrollChild)

		ls:SetPoint("TOPLEFT")
		ls:SetPoint("BOTTOMRIGHT", a, "BOTTOMLEFT", 150, 0)
		rs:SetPoint("TOPLEFT", ls, "TOPRIGHT", 10, 0)
		rs:SetPoint("BOTTOMRIGHT")

		a.leftScrollChild:SetWidth(200)
		a.leftScrollChild:SetHeight(300)
		a.rightScrollChild:SetWidth(500)
		a.rightScrollChild:SetHeight(50)

		rs:HookScript("OnVerticalScroll", areaOnVerticalScroll)
		rs:SetScript("OnHorizontalScroll", areaOnHorizontalScroll)

		-- Header
		local htl = a:CreateTexture(nil, "BORDER")
		a.headerTopL = htl
		htl:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopInset")
		htl:SetTexCoord(0, 0.25, 0, 1)
		htl:SetPoint("BOTTOMLEFT", a, "TOPLEFT", 120, -24)
		htl:SetWidth(64)
		htl:SetHeight(64)

		local htr = a:CreateTexture(nil, "BORDER")
		a.headerTopR = htr
		htr:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopInset")
		htr:SetTexCoord(0.75, 1, 0, 1)
		htr:SetPoint("BOTTOMRIGHT", a, "TOPRIGHT", 30, -24)
		htr:SetWidth(64)
		htr:SetHeight(64)

		local hmid = a:CreateTexture(nil, "BORDER")
		a.headerMid = hmid
		hmid:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopInset")
		hmid:SetTexCoord(0.0625, 0.9375, 0, 1)
		hmid:SetPoint("TOPLEFT", htl, "TOPRIGHT")
		hmid:SetPoint("BOTTOMRIGHT", htr, "BOTTOMLEFT")

		rs:EnableMouseWheel(true)
		rs:EnableMouse(true)
		rs:SetScript("OnMouseDown", areaOnMouseDown)
		rs:SetScript("OnMouseUp", areaOnMouseUp)
		rs:SetScript("OnSizeChanged", areaOnSizeChanged)
		rs:SetScript("OnEnter", areaRightScrollOnEnter)
		rs:SetScript("OnLeave", areaRightScrollOnLeave)
		rs:SetScript("OnMouseWheel", areaOnMouseWheel)

		rs.OnMouseDown = areaOnMouseDown
		rs.OnMouseUp = areaOnMouseUp

		-- Scale controls
		local sc = CreateFrame("Frame", nil, parent)
		a.scaleFrame = sc
		sc:SetPoint("BOTTOMLEFT", a, "TOPLEFT", 0, 10)
		sc:SetPoint("TOPRIGHT", htl, "TOPLEFT", -5, 0)

		sc.upButton = module:CreateArrowButton(sc, "up")
		sc.downButton = module:CreateArrowButton(sc, "down")
		sc.upButton:SetScript("OnClick", areaScaleUpOnClick)
		sc.downButton:SetScript("OnClick", areaScaleDownOnClick)

		sc.upButton:SetPoint("TOPRIGHT")
		sc.downButton:SetPoint("TOP", sc.upButton, "BOTTOM")
		sc.text = sc:CreateFontString(nil, "OVERLAY", "GameFontNormal")
		sc.text:SetPoint("RIGHT", sc.upButton, "BOTTOMLEFT", -4, 0)

		-- Functions
		a.SelectMobKey			= areaSelectMobKey
		a.OnClose				= areaOnClose
		a.ClearDisplay			= areaClearDisplay
		a.SetMode				= areaSetMode
		a.Init					= areaInit
		a.BuildSpellNameList	= areaBuildSpellNameList
		a.Populate				= areaPopulate
		a.ShowScroll			= areaShowScroll
		a.ShowScale				= areaShowScale
		a.ShowCursorLocation	= areaShowCursorLocation
		a.HideCursorLocation	= areaHideCursorLocation
		a.RemoveHighlight		= areaRemoveHighlight
		a.AddSpell				= areaAddSpell
		a.CreateSpellLine		= areaCreateSpellLine
		a.CreateSlidingIndicator = areaCreateSlidingIndicator
		a.Zoom					= areaZoom
		a.ShowDPS				= areaShowDPS

		return a
	end
end

do
	-- dps.SetFight
	local function dpsSetFight(self, fight)
		self.fight = fight
		self:CalcDPS()
		self:DrawDPS()
	end

	-- dps.OnHide
	local function dpsOnHide(self)
		self.fight = nil
		self.dmgAmounts = del(self.dmgAmounts)
		self.maxDamage = nil
		self:Cleanup()
	end

	-- dps.Cleanup
	local function dpsCleanup(self)
		if (self.texList) then
			for i,tex in ipairs(self.texList) do
				self:ReleaseDPSBlock(tex)
			end
			self.texList = del(self.texList)
		end
	end

	local blocks = {}
	local function dpsCreateDPSBlock(self, parent)
		local tex = parent:CreateTexture(nil, "OVERLAY")
		tex:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
		return tex
	end

	local function dpsAquireDPSBlock(self, parent)
		local tex = tremove(blocks) or self:CreateDPSBlock(parent)
		tex:SetParent(parent)
		tex:Show()
		return tex
	end

	local function dpsReleaseDPSBlock(self, tex)
		tex:ClearAllPoints()
		tex:SetParent(UIParent)
		tex:Hide()
		tex:SetHeight(0)
		tinsert(blocks, tex)
	end

	-- dps.DrawDPS
	local function dpsDrawDPS(self)
		self:Cleanup()
		if (not self.dmgAmounts) then
			return
		end

		local endSec = floor(self.fight.combatEnd - self.fight.combatStart + 1)
		local frame = self:GetParent()
		local area = frame.area
		local scale = area.scale
		local oneSecWidth = scale
		local height = self.height - 1

		self.texList = new()
		for i = 1,endSec do
			local amount = self.dmgAmounts[i]
			if (amount and amount > 0) then
				local tex = self:AquireDPSBlock(self.scrollChild)
				tinsert(self.texList, tex)

				tex:SetWidth(oneSecWidth)
				tex:SetHeight(amount / self.maxDamage * height)
				tex:SetPoint("BOTTOMLEFT", oneSecWidth * i, 0)
			end
		end

		self.scroll:SetHorizontalScroll(area.rightScroll:GetHorizontalScroll())
	end

	-- dps.CalcDPS
	local function dpsCalcDPS(self)
		local dps = self.fight.dps
		if (dps) then
			self.dmgAmounts = del(self.dmgAmounts)
			self.dmgAmounts = new()
			self.maxDamage = 0

			local area = self:GetParent().area
			for mobKey,data in pairs(dps) do
				if (area.mobKeys[mobKey]) then
					local temp = new(strsplit(",", data))
					for i,dmg in ipairs(temp) do
						self.dmgAmounts[i] = max(self.dmgAmounts[i] or 0, tonumber(dmg))
						self.maxDamage = max(self.maxDamage or 0, self.dmgAmounts[i])
					end
					del(temp)
				end
			end
		end
	end

	-- CreateDPSFrame
	function module:CreateDPSFrame(parent)
		local dps = CreateFrame("Frame", nil, parent)

		local scrollName = self:GenericSuffixName(parent, "ScrollFrame")
		local scroll = CreateFrame("ScrollFrame", scrollName, dps)
		dps.scroll = scroll
		dps.height = 100

		local scrollChild = CreateFrame("Frame", nil, scroll)
		dps.scrollChild = scrollChild
		scroll:SetScrollChild(scrollChild)

		scrollChild:SetWidth(400)
		scrollChild:SetHeight(dps.height - 1)

		local offset = parent.area.rightScroll:GetLeft() - parent.area.leftScroll:GetLeft()

		scroll:SetPoint("TOPLEFT", offset, 0)
		scroll:SetPoint("BOTTOMRIGHT")

	    dps.SetFight		= dpsSetFight
	    dps.OnHide			= dpsOnHide
	    dps.Cleanup			= dpsCleanup
	    dps.DrawDPS			= dpsDrawDPS
	    dps.CalcDPS			= dpsCalcDPS
		dps.CreateDPSBlock	= dpsCreateDPSBlock
		dps.AquireDPSBlock	= dpsAquireDPSBlock
		dps.ReleaseDPSBlock	= dpsReleaseDPSBlock

		return dps
	end
end

do
	local list = {}
	local listNames = {}

	-- AquireAreaTexture
	function module:AquireAreaTexture(frame, names)
		local tex
		if (names) then
			tex = tremove(listNames) or tremove(list) or self:CreateAreaTexture(frame)
		else
			tex = tremove(list) or tremove(listNames) or self:CreateAreaTexture(frame)
		end
		tex:SetParent(frame)
		if (tex.ownerText) then
			tex.ownerText:SetParent(frame)
		end
		tex:Show()
		return tex
	end

	-- ReleaseAreaTexture
	function module:ReleaseAreaTexture(tex)
		tex:ClearAllPoints()
		tex:Hide()
		tex:SetParent(UIParent)
		tex.spellId		= nil
		tex.realSpellId	= nil
		tex.sourceName	= nil
		tex.timeStart	= nil
		tex.timeEnd		= nil

		if (tex.ownerText) then
			tex.ownerText:Hide()
			tinsert(listNames, tex)
		else
			tinsert(list, tex)
		end
	end

	-- SetOwnerText
	local function setOwnerText(self, name, r, g, b)
		-- For some reason this will bump the position of the textures over a little
		-- Even the comment out code which anchors to the ScrollChild instead of texture will do exactly the same

		if (not self.ownerText) then
			self.ownerText = self:GetParent():CreateFontString(nil, "OVERLAY", "GameFontNormalSmall")
			--self.ownerText:SetHeight(20)
			self.ownerText:SetPoint("LEFT", self, "LEFT")
		else
			self.ownerText:Show()
		end
		self.ownerText:SetText(name)
		self.ownerText:SetTextColor(r or 0, g or 0, b or 0)

		--local point, _, rel, x, y = self:GetPoint(1)
		--self.ownerText:SetPoint(point, self:GetParent(), rel, x, y)
	end

	-- CreateAreaTexture
	function module:CreateAreaTexture(frame)
		local tex = frame:CreateTexture(nil, "OVERLAY")
		tex:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
		tex:SetVertexColor(0, 0.8, 0)
		tex:SetHeight(20)
		tex.SetOwnerText = setOwnerText
		return tex
	end
end

do
	local function frameDragStop(self)
		self:StopMovingOrSizing()
		Utopia:SavePosition(self)
	end

	local function frameGripDragStart(self)
		self:GetParent():StartSizing()
	end

	local function frameGripDragStop(self)
		self:GetParent():StopMovingOrSizing()
		Utopia:SavePosition(self:GetParent())
	end

	-- MovableFrame
	function module:MovableFrame(frame)
		tinsert(UISpecialFrames, frame:GetName())

		frame:SetToplevel(true)
		frame:SetMovable(true)
		frame:SetResizable(true)
		frame:EnableMouse(true)
		frame:RegisterForDrag("LeftButton")

		frame.grip = CreateFrame("Frame", nil, frame)
		frame.grip:SetPoint("BOTTOMRIGHT", -4, 7)
		frame.grip:SetWidth(20)
		frame.grip:SetHeight(20)
		frame.grip.tex = frame.grip:CreateTexture(nil, "OVERLAY")
		frame.grip.tex:SetAllPoints()
		frame.grip.tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\ResizeGrip")
		frame.grip.tex:SetBlendMode("ADD")
		frame.grip.hl = frame.grip:CreateTexture(nil, "HIGHLIGHT")
		frame.grip.hl:SetAllPoints()
		frame.grip.hl:SetTexture("Interface\\Addons\\Utopia\\Textures\\ResizeGrip")
		frame.grip.hl:SetBlendMode("ADD")

		frame:SetScript("OnDragStart", frame.StartMoving)
		frame:SetScript("OnDragStop", frameDragStop)

		frame.grip:EnableMouse(true)
		frame.grip:RegisterForDrag("LeftButton")
		frame.grip:SetScript("OnDragStart", frameGripDragStart)
		frame.grip:SetScript("OnDragStop", frameGripDragStop)

		Utopia:RestorePosition(frame)
	end
end

-- CreateArrowButton
function module:CreateArrowButton(parent, which)
	local template
	if (which == "up") then
		template = "UIPanelScrollUpButtonTemplate"
	elseif (which == "down") then
		template = "UIPanelScrollDownButtonTemplate"
	end

	local b = CreateFrame("Button", nil, parent, template)
	return b
end

-- Button
function module:Button(parent, text, click, anchor, x, y, relFrame, rel)
--@debug@
	assert(type(parent) == "table" and parent.GetFrameType)
	assert(type(text) == "string")
	assert(type(click) == "function")
	assert(type(anchor) == "string")
	assert(not x or type(x) == "number")
	assert(not y or type(y) == "number")
	assert(not relFrame or (type(relFrame) == "table" and relFrame.GetFrameType))
	assert(not rel or type(rel) == "string")
--@debug@

	local b = CreateFrame("Button", nil, parent, "OptionsButtonTemplate")
	b.text = b:GetRegions()
	b.text:SetAllPoints(b)			-- Makes the text part (first region) fit all over button, instead of just centered and fuxed when scaled
	b:SetScript("OnClick", click)
	b:SetText(text)
	b:SetWidth(max(100, b.text:GetStringWidth() + 25))
	b:SetHeight(26)
	b:SetPoint(anchor, relFrame or parent, rel or anchor, x, y)
	return b
end

do
	-- checkboxOnClick
	local function checkboxOnClick(self)
		self.set(self, self:GetChecked())
	end

	-- checkboxOnShow
	local function checkboxOnShow(self)
		self:SetChecked(self.get(self))
	end

	local function checkboxOnEnter(self)
		self.text:SetTextColor(1, 1, 1)
	end

	local function checkboxOnLeave(self)
		self.text:SetTextColor(1, 0.82, 0)
	end

	local function checkboxSetEnabled(self, enabled)
		if (enabled) then
			self:Enable()
			self.text:SetTextColor(1, 0.82, 0)
		else
			self:Disable()
			self.text:SetTextColor(0.5, 0.5, 0.5)
		end
	end

	-- CheckBox
	function module:CheckBox(parent, text, get, set, anchor, x, y, relFrame, rel)
--@debug@
		assert(type(parent) == "table" and parent.GetFrameType)
		assert(type(text) == "string")
		assert(type(get) == "function")
		assert(type(set) == "function")
		assert(type(anchor) == "string")
		assert(not x or type(x) == "number")
		assert(not y or type(y) == "number")
		assert(not relFrame or (type(relFrame) == "table" and relFrame.GetFrameType))
		assert(not rel or type(rel) == "string")
--@debug@

		local b = CreateFrame("CheckButton", nil, parent, "OptionsCheckButtonTemplate")
		local temp = new(b:GetRegions())
		for i,reg in ipairs(temp) do
			if (reg:GetObjectType() == "FontString") then
				b.text = reg
				break
			end
		end
		assert(b.text)
		b:SetScript("OnClick", checkboxOnClick)
		b:SetScript("OnShow", checkboxOnShow)
		b:SetScript("OnEnter", checkboxOnEnter)
		b:SetScript("OnLeave", checkboxOnLeave)
		b.SetEnabled = checkboxSetEnabled

		b.set = set
		b.get = get
		b.text:SetText(text)
		b:SetHeight(26)
		b:SetWidth(26)
		b:SetPoint(anchor, relFrame or parent, rel or anchor, x, y)
		b:SetHitRectInsets(0, -(b.text:GetStringWidth() + 5), 0, 0)
		return b
	end
end

-- ApplyBackground
function module:ApplyBackground(frame)
	frame.bg = {}

	frame:SetMinResize(470, 260)

	frame.title = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	frame.title:SetPoint("TOPLEFT", 10, -15)
	frame.title:SetPoint("BOTTOMRIGHT", frame, "TOPRIGHT", -10, -35)

	frame.bg.corner = frame:CreateTexture(nil, "OVERLAY")
	frame.bg.corner:SetTexture("Interface\\Addons\\Utopia\\Textures\\UtopiaCornerIcon")
	frame.bg.corner:SetWidth(59)
	frame.bg.corner:SetHeight(59)
	frame.bg.corner:SetPoint("TOPLEFT", 9, -6)

	frame:SetScript("OnSizeChanged",
		function(self, x, y)
			self:ArrangeTextures()
		end
	)

	function frame:ArrangeTextures()
		local h, w = self:GetHeight(), self:GetWidth()
		local rows = max(2, ceil(h / 256))
		local cols = max(2, ceil(w / 256))

		if (self.backgroundTextures) then
			for i,row in ipairs(self.backgroundTextures) do
				for j,tex in ipairs(row) do
					tex:Hide()
				end
			end
		else
			self.backgroundTextures = {}
		end

		local shrinkW = w < 512 and w / 2
		local shrinkH = h < 512 and h / 2
		local prevRow, prev

		for row = 1,rows do
			local rowTextures = self.backgroundTextures[row]
			if (not rowTextures) then
				rowTextures = {}
				self.backgroundTextures[row] = rowTextures
			end

			for col = 1,cols do
				local tex = rowTextures[col]
				if (not tex) then
					tex = frame:CreateTexture(nil, "BACKGROUND")
					rowTextures[col] = tex
				end
				tex:ClearAllPoints()
				tex:Show()

				if (col == 1 and row == 1) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopLeft")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetTexCoord(0, (shrinkW or 256) / 256, 0, (shrinkH or 256) / 256)
					tex:SetPoint("TOPLEFT")

				elseif (row == 1 and col < cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-Top")
					if (col == cols - 1) then
						local sw = w % 256
						tex:SetWidth(sw)
						tex:SetTexCoord(0, sw / 256, 0, (shrinkH or 256) / 256)
					else
						tex:SetWidth(256)
						tex:SetTexCoord(0, 1, 0, (shrinkH or 256) / 256)
					end
					tex:SetPoint("TOPLEFT", prev, "TOPRIGHT")
					tex:SetHeight(shrinkH or 256)
					
				elseif (row == 1 and col == cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-TopRight")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 0, (shrinkH or 256) / 256)
					tex:SetPoint("TOPRIGHT")

				elseif (row == rows and col == 1) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomLeft")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetTexCoord(0, (shrinkW or 256) / 256, 1 - (shrinkH or 256) / 256, 1)
					tex:SetPoint("BOTTOMLEFT")

				elseif (row == rows and col < cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-Bottom")
					if (col == cols - 1) then
						local sw = w % 256
						tex:SetWidth(sw)
						tex:SetTexCoord(0, sw / 256, 1 - (shrinkH or 256) / 256, 1)
					else
						tex:SetWidth(256)
						tex:SetTexCoord(0, 1, 1 - (shrinkH or 256) / 256, 1)
					end
					tex:SetPoint("BOTTOMLEFT", prev, "BOTTOMRIGHT")
					tex:SetHeight(shrinkH or 256)

				elseif (row == rows and col == cols) then
					tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomRight")
					tex:SetWidth(shrinkW or 256)
					tex:SetHeight(shrinkH or 256)
					tex:SetPoint("BOTTOMRIGHT")
					tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 1 - (shrinkH or 256) / 256, 1)

				elseif (row > 1 and row < rows) then
					if (col == 1) then
						tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomLeft")
						tex:SetTexCoord(0, 1, 0, 0.8)
						if (row == rows - 1) then
							local sh = h % 256
							tex:SetHeight(sh)
							tex:SetTexCoord(0, (shrinkW or 256) / 256, 0, 0.8 / 256 * sh)
						else
							tex:SetHeight(256)
							tex:SetTexCoord(0, (shrinkW or 256) / 256, 0, 0.8)
						end
						tex:SetWidth(shrinkW or 256)
						tex:SetPoint("TOPLEFT", prevRow, "BOTTOMLEFT")

					elseif (col == cols) then
						tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomRight")
						tex:SetTexCoord(0, 1, 0, 0.8)
						if (row == rows - 1) then
							local sh = h % 256
							tex:SetHeight(sh)
							tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 0, 0.8 / 256 * sh)
						else
							tex:SetHeight(256)
							tex:SetTexCoord(1 - (shrinkW or 256) / 256, 1, 0, 0.8)
						end
						tex:SetWidth(shrinkW or 256)
						tex:SetPoint("TOPLEFT", prev, "TOPRIGHT")

					elseif (col < cols) then
						tex:SetTexture("Interface\\Addons\\Utopia\\Textures\\UpTime-BottomRight")
						local sw = w % 256
						local sh = h % 256
						local height, width = 256, 256
						local tch = 0.8
						local tcw = 0.8

						if (row == rows - 1) then
							height = sh
							tch = 0.8 / 256 * sh
						end
						if (col == cols - 1) then
							width = sw
							tcw = 0.8 / 256 * sw
						end

						tex:SetWidth(width)
						tex:SetHeight(height)
						tex:SetTexCoord(0, tcw, 0, tch)
						tex:SetPoint("TOPLEFT", prev, "TOPRIGHT")
					end

				end

				if (col == 1) then
					prevRow = tex
				end
				prev = tex
			end
		end
	end

	frame.closeButton = self:CloseButton(frame)
end

-- DescribeDate
function module:DescribeDate(describe)
	local today = date("%x", time())
	local yesterday = date("%x", time() - 24*60*60)
	local week = time() - 7*24*60*60
	local when = date("%x", describe)
	local age
	if (when == today) then
		when = L["Today"]
		age = 1
	elseif (when == yesterday) then
		when = L["Yesterday"]
		age = 2
	elseif (describe > week) then
		when = date("%A", describe)
		age = 3
	else
		when = date("%X", describe)
		age = 4
	end
	return when, age
end

do
	-- OnShow
	local function mainlistOnShow(self)
		self.offset, self.offset2 = 0, 0
		self:Populate()
		self.list:Select(1)
	end

	-- OnHide
	local function mainlistOnHide(self)
		self.list:Select()
	end

	-- list.OnVerticalScroll
	local function mainlistOnVerticalScroll(self, offset)
		self:GetParent().offset = offset
		self:GetParent():Populate()
	end

	-- list2.OnVerticalScroll
	local function mainlistOnVerticalScroll2(self, offset)
		self:GetParent().offset2 = offset
		self:GetParent():Populate2()
	end

	-- list.OnClick
	local function mainlistOnClick(self)
		self:GetParent():GetParent():Populate2(self.key)
	end

	-- list2.OnClick
	local function mainlistOnClick2(self)
		module:OpenView(self.key)
	end

	-- list.OnSizeChanged
	local function mainlistOnSizeChanged(self)
		self:GetParent():Populate()
	end

	-- list2.OnSizeChanged
	local function mainlistOnSizeChanged2(self)
		self:GetParent():Populate2()
	end

	local ageColours = {"|cFF90FF90", "|cFFFFFF80", "|cFFFFA020", "|cFFFF4040"}

	-- Populate
	local function mainlistPopulate(self)
		local lines = floor(self.list:GetHeight() / 16)
		local last
		local line, total = 0, 0

		if (db.history) then
			local temp = new()

			for i = #db.history,1,-1 do
				local info = db.history[i]
				local when, age = module:DescribeDate(info.combatStart)
				local colour = ageColours[age]

				local key = format("%s: %s", info.player, when)
				if (key ~= last) then
					if (not temp[key]) then
						temp[key] = true

						last = key
						line = line + 1

						if (line >= self.offset and line <= lines + self.offset) then
							local b = self.list.lines[line]
							b:SetText(format("%s: %s%s", info.player, colour, when))
							b.key = key
							b:Show()
						end
					end
				end
			end

			del(temp)
		end
		while (line < #self.list.lines) do
			line = line + 1
			self.list.lines[line]:Hide()
		end

		self.list.scroll:Update(total)
	end

	-- Populate2
	local function mainlistPopulate2(self, key)
		if (not db.history) then
			return
		end

		self.key = key or self.key

		local lines = floor(self.list2:GetHeight() / 16)
		local line, total = 0, 0

		for i = #db.history,1,-1 do
			local info = db.history[i]
			local when = module:DescribeDate(info.combatStart)

			local display = format("%s: %s", info.player, when)
			if (display == self.key) then
				total = total + 1

				if (total > self.offset2 and total <= lines + self.offset2) then
					line = line + 1
					local b = self.list2.lines[line]

					-- Pick an appropriate mob name to display
					local _, mobname = next(info.mobs)
					for j,name in pairs(info.mobs) do
						if (bossid[j]) then
							mobname = name			-- But use a boss' name if there's one
							break
						end
					end

					local dur = info.combatEnd - info.combatStart
					if (dur > 60) then
						dur = format(L["%.1f mins"], dur / 60)
					else
						dur = format(L["%d secs"], dur)
					end
						
					b:SetFormattedText("%s - %s - %s", date("%H:%M", info.combatStart), dur, mobname or "")
					b.key = i
					b:Show()
				end
			end
		end
		while (line < #self.list2.lines) do
			line = line + 1
			self.list2.lines[line]:Hide()
		end

		self.list2.scroll:Update(total)
	end

	-- CreateMainList
	function module:CreateMainList()
		local frame = CreateFrame("Frame", "UtopiaUpTimeList", UIParent)
		frame:Hide()
		frame:SetWidth(512)
		frame:SetHeight(512)
		frame:SetScale(0.8)
		frame.offset, frame.offset2 = 0, 0

		self:ApplyBackground(frame)
		self:MovableFrame(frame)

		frame:SetScript("OnHide", mainlistOnHide)
		frame:SetScript("OnShow", mainlistOnShow)

		frame.title:SetText(L["Fight Listing"])

		frame.list = self:CreateListFrame(frame, true)
		frame.list:SetPoint("TOPLEFT", 24, -82)
		frame.list:SetPoint("BOTTOMRIGHT", frame, "BOTTOMLEFT", 200, 42)

		frame.list.OnClick = mainlistOnClick
		frame.Populate = mainlistPopulate
		frame.list.OnVerticalScroll = mainlistOnVerticalScroll
		frame.list:SetScript("OnSizeChanged", mainlistOnSizeChanged)

		frame.bar = self:CreateVerticalBar(frame.list, "RIGHT")

		frame.list2 = self:CreateListFrame(frame)
		frame.list2:SetPoint("TOPLEFT", frame.list, "TOPRIGHT", 12, 0)
		frame.list2:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -40, 42)

		frame.list2.OnClick = mainlistOnClick2
		frame.Populate2 = mainlistPopulate2
		frame.list2.OnVerticalScroll = mainlistOnVerticalScroll2
		frame.list2:SetScript("OnSizeChanged", mainlistOnSizeChanged2)

		return frame
	end
end

-- GenericSuffixName
function module:GenericSuffixName(frame, suffix)
	local name, prefix
	if (frame:GetName()) then
		prefix = frame:GetName()
	else
		prefix = "Utopia"
	end
	if (not self.suffix) then
		self.suffix = {}
	end
	self.suffix[suffix] = (self.suffix[suffix] or 0) + 1
	return format("%s%s%d", prefix, suffix, self.suffix[suffix])
end

do
	local function onHide(self)
		self:GetParent():Hide()
	end
	function module:CloseButton(frame)
		local name = self:GenericSuffixName(frame, "CloseButton")
		local b = CreateFrame("Button", name, frame, "UIPanelCloseButton")
		b:SetScript("OnClick", onHide)
		b:SetPoint("TOPRIGHT", -3, -8)
		return b
	end
end

do
	-- OnClick
	local function entryOnClick(self)
		local p = self:GetParent()
		if (p.selectable) then
			local current = self:GetID() + p.offset
			if (p.multiselect and IsControlKeyDown()) then
				p.selected[current] = not p.selected[current] or nil

			elseif (p.multiselect and IsShiftKeyDown()) then
				for i = p.focus,current,(p.focus < current and 1 or -1) do
					local line = rawget(p.lines, i - p.offset)
					if (line) then
						line:LockHighlight()
					end
					p.selected[i] = true
				end
			else
				for i,line in pairs(p.lines) do
					line:UnlockHighlight()
				end
				wipe(p.selected)
				p.selected[current] = true
			end

			if (p.selected[current]) then
				self:LockHighlight()
			else
				self:UnlockHighlight()
			end

			p.focus = current
		end
		p:DoOnClick(self)
	end

	-- DoOnClick
	local function listDoOnClick(self, from)
		if (self.OnClick) then
			if (self.multiselect) then
				local temp = new()
				for id in pairs(self.selected) do
					tinsert(temp, id)
				end
				sort(temp)
				self.OnClick(from, unpack(temp))
				del(temp)
			else
				self.OnClick(from, (self.selectable and self.selected and next(self.selected)))
			end
		end
	end

	-- Select
	local function listSelect(self, index)
		if (self.selectable) then
			if (not self.selected) then
				self.selected = {}
			else
				wipe(self.selected)
			end
			if (index) then
				self.selected[index] = true
				local line = rawget(self.lines, index - self.offset)
				if (line) then
					entryOnClick(line)
				end
			else
				for i,line in pairs(self.lines) do
					line:UnlockHighlight()
				end
			end
		end
	end

	-- GetSelected
	local function listGetSelected(self)
		if (self.selectable) then
			if (self.multiselect) then
				local temp = new()
				for id in pairs(self.selected) do
					tinsert(temp, id)
				end
				sort(temp)
				return unpack(temp)
			else
				return (next(self.selected))
			end
		end
	end

	-- OnUpdateDragging
	local function listOnUpdateDragging(self, elapsed)
		if (MouseIsOver(self)) then
			local dragSelectEnd
			for i = 1,#self.lines do
				local line = rawget(self.lines, i)
				if (line:IsShown() and MouseIsOver(line)) then
					dragSelectEnd = line:GetID()
					break
				end
			end

			if (dragSelectEnd) then
				for i,line in pairs(self.lines) do
					line:UnlockHighlight()
				end
				wipe(self.selected)

				for i = self.dragSelectStart,dragSelectEnd,(self.dragSelectStart < dragSelectEnd and 1 or -1) do
					local line = rawget(self.lines, i - self.offset)
					if (line) then
						line:LockHighlight()
					end
					self.selected[i] = true
				end
			end
		end
	end

	-- OnDragStart
	local function entryOnDragStart(self)
		local p = self:GetParent()
		p.dragSelectStart = self:GetID() + p.offset
		p:SetScript("OnUpdate", listOnUpdateDragging)
	end

	-- OnDragStop
	local function entryOnDragStop(self)
		local p = self:GetParent()
		p:SetScript("OnUpdate", nil)
		self.dragSelectStart = nil
		p:DoOnClick(self)
	end

	-- CreateListEntry
	local function CreateListEntry(parent, index, lines)
		local b = CreateFrame("Button", nil, parent)
		b:SetNormalFontObject(GameFontNormal)
		b:SetHighlightFontObject(GameFontHighlight)
		b:SetDisabledFontObject(GameFontDisable)
		b:SetHighlightTexture("Interface\\QuestFrame\\UI-QuestTitleHighlight")
		b:SetText(" ") -- Required for GetFontString to work
		b.text = b:GetFontString()
		b.text:SetAllPoints()
		b.text:SetJustifyH("LEFT")

		b:SetID(index)
		if (index == 1) then
			b:SetPoint("TOPLEFT")
			b:SetPoint("BOTTOMRIGHT", parent, "TOPRIGHT", 0, -16)
		else
			b:SetPoint("TOPLEFT", rawget(lines, index - 1), "BOTTOMLEFT")
			b:SetPoint("BOTTOMRIGHT", rawget(lines, index - 1), "BOTTOMRIGHT", 0, -16)
		end
		b:SetScript("OnClick", entryOnClick)

		if (parent.multiselect) then
			b:RegisterForDrag("LeftButton")
			b:SetScript("OnDragStart", entryOnDragStart)
			b:SetScript("OnDragStop", entryOnDragStop)
		end

		return b
	end

	-- CreateListFrame
	function module:CreateListFrame(parent, selectable, multiselect)
		local name = self:GenericSuffixName(parent, "List")
		local frame = CreateFrame("Frame", name, parent)
		frame.selectable = selectable
		frame.multiselect = multiselect
		frame.offset, frame.focus = 0, 0
		if (selectable or multiselect) then
			frame.selected = {}
		end

		frame.lines = setmetatable({}, {
			__index = function(self, index)
				assert(type(index) == "number")
				local b = CreateListEntry(frame, index, self)
				self[index] = b
				return b
			end
		})

		frame.scroll = self:CreateScrollFrame(frame)
		frame.Select = listSelect
		frame.GetSelected = listGetSelected
		frame.DoOnClick = listDoOnClick

		return frame
	end
end

do
	local function scrollUpdate(self, total)
		local parent = self:GetParent()
		if (not self.parentWidth) then
			self.parentWidth = parent:GetWidth()
		end

		local visibleLines = floor(self:GetParent():GetHeight() / 16)
		if (FauxScrollFrame_Update(self, total, visibleLines, 16)) then
			parent:SetWidth(self.parentWidth - 21)
			self:Show()
		else
			parent:SetWidth(self.parentWidth)
			self:Hide()
		end
	end

	local function onverticalscroll(self)
		if (self:GetParent().OnVerticalScroll) then
			self:GetParent():OnVerticalScroll(self.offset)
		end
	end

	local function scrollOnVerticalScroll(self, offset)
		self.offset = offset
    	FauxScrollFrame_OnVerticalScroll(self, offset, 16, onverticalscroll)
	end

	-- CreateScrollFrame
	function module:CreateScrollFrame(frame)
		local name = self:GenericSuffixName(frame, "ScrollBar")
		local scroll = CreateFrame("ScrollFrame", name, frame, "FauxScrollFrameTemplate")
		scroll:Hide()
		scroll:SetPoint("TOPRIGHT", 1, 0)
		scroll:SetPoint("BOTTOMLEFT", 0, 2)
		scroll:SetScript("OnVerticalScroll", scrollOnVerticalScroll)
		scroll.Update = scrollUpdate

		scroll.tex1 = scroll:CreateTexture(nil, "BORDER")
		scroll.tex1:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-ScrollBar")
		scroll.tex1:SetWidth(31)
		scroll.tex1:SetHeight(64)
		scroll.tex1:SetPoint("TOPLEFT", scroll, "TOPRIGHT", -2, 3)
		scroll.tex1:SetTexCoord(0, 0.484375, 0, 0.25)

		scroll.tex2 = scroll:CreateTexture(nil, "BORDER")
		scroll.tex2:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-ScrollBar")
		scroll.tex2:SetWidth(31)
		scroll.tex2:SetHeight(106)
		scroll.tex2:SetPoint("BOTTOMLEFT", scroll, "BOTTOMRIGHT", -2, -2)
		scroll.tex2:SetTexCoord(0.515625, 1, 0, 0.4140625)

		scroll.tex3 = scroll:CreateTexture(nil, "BORDER")
		scroll.tex3:SetTexture("Interface\\PaperDollInfoFrame\\UI-Character-ScrollBar")
		scroll.tex3:SetPoint("TOPLEFT", scroll.tex1, "BOTTOMLEFT")
		scroll.tex3:SetPoint("BOTTOMRIGHT", scroll.tex2, "TOPRIGHT")
		scroll.tex3:SetTexCoord(0.515625, 1, 0, 0.3)

		return scroll
	end
end

do
	-- OnUpdate
	local function barOnUpdate(self, elapsed)
		local xCursor, yCursor = GetCursorPosition()
		xCursor, yCursor = xCursor / self:GetEffectiveScale(), yCursor / self:GetEffectiveScale()
		local xFrame, yFrame = self:GetCenter()

		local parentFrame = self:GetParent()
		if (self.vertical) then
			if (xCursor < xFrame - 1 or xCursor > xFrame + 1) then
				local point, parent, relpoint, x, y = parentFrame:GetPoint(2)
				x = x - (xFrame - xCursor)
				parentFrame:SetPoint(point, parent, relpoint, x, y)
			end
		else
			if (yCursor < yFrame - 1 or xCursor > yFrame + 1) then
				local point, parent, relpoint, x, y = parentFrame:GetPoint(2)
				y = y - (yFrame - yCursor)
				parentFrame:SetPoint(point, parent, relpoint, x, y)
			end
		end
	end

	-- OnDragStart
	local function barOnDragStart(self)
		self:SetScript("OnUpdate", barOnUpdate)
	end

	-- OnDragStop
	local function barOnDragStop(self)
		self:SetScript("OnUpdate", nil)
	end

	-- CreateVerticalBar
	function module:CreateVerticalBar(frame, where, draggable)
		assert(where == "LEFT" or where == "RIGHT")
		local bars = CreateFrame("Frame", nil, frame)
		bars.bars = {}
		bars.vertical = true
		local notWhere = where == "LEFT" and "RIGHT" or "LEFT"

		bars:SetPoint("TOP"..notWhere, frame, "TOP"..where, 0, 2)
		bars:SetPoint("BOTTOM"..where, frame, "BOTTOM"..where, where == "RIGHT" and 8 or -8, 0)

		for i = 1,3 do
			local bar = bars:CreateTexture(nil, "OVERLAY")
			bar:SetTexture("Interface\\FriendsFrame\\UI-ChannelFrame-VerticalBar")
			tinsert(bars.bars, bar)

			if (i == 1) then
				bar:SetPoint("TOPLEFT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0, 0.171875, 0, 0.171875)
			elseif (i == 2) then
				bar:SetPoint("BOTTOMLEFT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0.82, 0.98, 0.798, 0.9609375)
			else
				bar:SetPoint("TOPLEFT", bars.bars[1], "BOTTOMLEFT")
				bar:SetPoint("BOTTOMRIGHT", bars.bars[2], "TOPRIGHT")
				bar:SetTexCoord(0.39, 0.6, 0, 1)
			end
		end

		if (draggable) then
			bars.highlight = bars:CreateTexture(nil, "HIGHLIGHT")
			bars.highlight:SetPoint("TOPLEFT", 3, 0)
			bars.highlight:SetPoint("BOTTOMRIGHT", -3, 0)
			bars.highlight:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
			bars.highlight:SetVertexColor(1, 1, 1, 0.5)

			bars:EnableMouse(true)
			bars:RegisterForDrag("LeftButton")
			bars:SetScript("OnDragStart", barOnDragStart)
			bars:SetScript("OnDragStop", barOnDragStop)
		end

		return bars
	end

	-- CreateHorizontalBar
	function module:CreateHorizontalBar(frame, where, draggable)
		assert(where == "TOP" or where == "BOTTOM")
		local bars = CreateFrame("Frame", nil, frame)
		bars.bars = {}
		bars.horizontal = true
		bars.where = where
		local notWhere = where == "TOP" and "BOTTOM" or "TOP"

		bars:SetPoint(notWhere.."LEFT", frame, where.."LEFT", -4, 0)
		bars:SetPoint(where.."RIGHT", frame, where.."RIGHT", 0, where == "TOP" and 8 or -8)

		for i = 1,3 do
			local bar = bars:CreateTexture(nil, "OVERLAY")
			bar:SetTexture("Interface\\ClassTrainerFrame\\UI-ClassTrainer-HorizontalBar")
			tinsert(bars.bars, bar)

			if (i == 1) then
				bar:SetPoint("TOPLEFT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0, 0.25, 0, 0.25)
			elseif (i == 2) then
				bar:SetPoint("TOPRIGHT")
				bar:SetWidth(8)
				bar:SetHeight(8)
				bar:SetTexCoord(0.09, 0.3, 0.25, 0.5)
			else
				bar:SetPoint("TOPLEFT", bars.bars[1], "TOPRIGHT")
				bar:SetPoint("BOTTOMRIGHT", bars.bars[2], "BOTTOMLEFT")
				bar:SetTexCoord(0.06, 1, 0, 0.25)
			end
		end

		if (draggable) then
			bars.highlight = bars:CreateTexture(nil, "HIGHLIGHT")
			bars.highlight:SetPoint("TOPLEFT", 0, -3)
			bars.highlight:SetPoint("BOTTOMRIGHT", 0, 3)
			bars.highlight:SetTexture("Interface\\Tooltips\\UI-Tooltip-Background")
			bars.highlight:SetVertexColor(1, 1, 1, 0.5)

			bars:EnableMouse(true)
			bars:RegisterForDrag("LeftButton")
			bars:SetScript("OnDragStart", barOnDragStart)
			bars:SetScript("OnDragStop", barOnDragStop)
		end

		return bars
	end
end

-- Toggle
function module:Toggle()
	if (not self.listFrame) then
		self.listFrame = self:CreateMainList()
	end
	if (self.listFrame:IsShown()) then
		self.listFrame:Hide()
	else
		self.listFrame:Show()
	end
end

-- OnInitialize
function module:OnInitialize()
	local defaults = { 
		profile = {
			bossesOnly = true,
			notifications = true,
			minimumDuration = 30,
			keep = 25,
			coloursExtreme = true,
		} 
	}

	-- If we're loading modular, then we'll store our DB associated with the module instead of Utopia proper
	-- Mostly, so I can easily check what the memory usage of this module's data will be like
	if (IsAddOnLoaded("Utopia_UpTime")) then
		self.db = LibStub("AceDB-3.0"):New("UtopiaUpTimeDB", defaults, "Default")
		self:MoveDataFromUtopiaDBtoLocalDB()
	else
		self.db = Utopia.db:RegisterNamespace("UpTime", defaults)
	end

	db = self.db.profile
end

local svList = {"history", "bossesOnly", "notifications", "minimumDuration", "keep"}

-- MoveDataFromUtopiaDBtoLocalDB
function module:MoveDataFromUtopiaDBtoLocalDB()
	if (Utopia.db.sv.namespaces.UpTime) then
		-- We moved from embeded to modular, move the settings over. (can't go the other way tho)
		local profiles = Utopia.db.sv.namespaces.UpTime.profiles
		local oldIntegratedDB = profiles and profiles.Default
		if (oldIntegratedDB and oldIntegratedDB.history and not self.db.profile.history) then
			for i,key in pairs(svList) do
				self.db.profile[key] = oldIntegratedDB[key]
				oldIntegratedDB[key] = nil
			end
		end
	end
end

-- MoveDataFromLocalDBtoUtopiaDB
function module:MoveDataFromLocalDBtoUtopiaDB()
	local ns = Utopia.db.sv.namespaces
	if (ns) then
		if (not ns.UpTime) then
			ns.UpTime = {profiles = {Default = {}}}
		end

		local newDB = ns.UpTime.profiles.Default

		for i,key in pairs(svList) do
			newDB[key] = self.db.profile[key]
			self.db.profile[key] = nil
		end
	end
end

-- OnProfileChanged
function module:OnProfileChanged(event, database, newProfileKey)
	db = database.profile
end

-- OnEnable
function module:OnEnable()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	self:RegisterEvent("PLAYER_REGEN_ENABLED")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("PLAYER_LEAVING_WORLD")
	self:RawHook("SetItemRef", true)

	if (InCombatLockdown()) then
		self:StartFighting()
	end

	FRIEND_FLAGS = COMBATLOG_OBJECT_AFFILIATION_MINE + COMBATLOG_OBJECT_AFFILIATION_PARTY + COMBATLOG_OBJECT_AFFILIATION_RAID
	ENEMY_FLAGS = COMBATLOG_OBJECT_REACTION_HOSTILE
	EXLCUDE_FLAGS = COMBATLOG_OBJECT_TYPE_OBJECT + COMBATLOG_OBJECT_TYPE_PLAYER

	db = self.db.profile
end

-- OnDisable
function module:OnDisable()
	self.record, self.linear, self.combatStart = nil, nil, nil
end
